#*************************************************************************************************************
# Name: BF2:Sandbox
# Author: Elton "Elxx" Muuga (http://elxx.net)
# Version: 0.5 (June 29, 2007)
# Website: http://sandboxmod.com
#*************************************************************************************************************
# Conditions and Limitations: You may use this file locally, on listening and dedicated servers.
# This software is provided "as is". In no event shall the author, authors or contributors be held
# liable for  any direct, indirect, incidental, special, exemplary, or consequential damages (including)
# but not limited to, procurement of substitute good or services; loss of use, data or profits; or
# business interruption) however caused and on any theory of liability, whether in contract, strict liability
# arising in any way out of the use of this software, even if advised of the possibility of such damage.
#*************************************************************************************************************
#
# It's probably not a good idea to poke around in here if you don't know what you're doing. Go to the
# settings folder and edit sandbox.ini if you're bored and looking for something to do.
#
#*************************************************************************************************************

import host
import bf2
import bf2.Timer
import bf2.PlayerManager
import math
import sys
import re
import os
import inspect
import sandboxUploader
import time
import ConfigParser
from socket import *
from bf2.stats.constants import *

class pData:
	def __init__(self):
		self.tracerDistance = 3
		self.objects = []
		self.objectTypes = []
		self.grabbed = []
		self.lastRotation = 0
		self.savedRotation = ()
		self.lastSpawned = []
		self.lastTemplate = ""
		self.tracerOn = 1
		self.tracerDistance = 8
		self.gridSnap = 0
		self.rotSnap = 5
		self.autoGrab = 1
		self.autoLock = 0
		self.numSpawned = [0,0]
		self.preGrab = []
		self.preGrabMode = 0
		self.lastAction = []
		self.spawnDelay = 0
		self.dupeDelay = 0
		self.saveDelay = 0
		self.grabMode = 3
		self.groups = [] # List of all the groups the user has created
		self.selectedGroup = [] # Newly spawned and added objects will be added to this group
		self.groupCreate = False # Set to True if new objects will be added to group
		self.grabbedGroupAnchor = 0 # The object of the grabbed group currently selected
		self.grabbedGroup = [] # Currently grabbed group, list of [obj, relative_position]
		self.lastGroup = []
		self.startMerge = False
		self.popMessage = ""

	def clearGrabbed(self):
		self.grabbed = []
		self.grabbedGroupAnchor = 0
		self.grabbedGroup = []

	def dropGroup(self):
		self.grabbedGroup = []
		self.grabbedGroupAnchor = 0
		self.lastGroup = []

# Grab Modes
# 1 = Original, own objects
# 2 = Original, all objects
# 3 = Extra-Accuracy, own objects
# 4 = Extra-Accuracy, all objects

# Undo Actions
# 1 = grab [object, original position, original rotation]
# 2 = drop [object, frozen]
# 3 = delete [object, original position] (disabled)
# 4 =

def ExceptionOutput():
	sys.stderr.write("\n" + "Exception Occured: " + str(sys.exc_info()[0]) + "\n")
	sys.stderr.write("Value: " + str(sys.exc_info()[1]) + "\n")
	sys.stderr.write("Line:" + str(readline(inspect.getfile(sys.exc_info()[2]),
							sys.exc_info()[2].tb_lineno)) + "\n")
	sys.stderr.write("Line #: " + str(sys.exc_info()[2].tb_lineno) + "\n")
	sys.stderr.write("File: " + str(inspect.getfile(sys.exc_info()[2])) + "\n" + "\n")
	#host.rcon_invoke("game.sayAll \"Exception Caught: " + str(sys.exc_info()[1]) + "\"")
	#host.rcon_invoke("game.sayAll \"Line " + str(sys.exc_info()[2].tb_lineno) + "\"")

def readline(filename, lineno):
	filen = re.sub('\\\\', '/', filename)
	file = open(filen, 'rU')
	lines = file.readlines()
	file.close()
	linen = lineno - 1
	line = re.sub('\s+', ' ', lines[linen])
	return line

SANDBOX_VERSION = "0.5"
SANDBOX_BUILD = "0630"

class sandboxCore:
	def __init__(self):
		self.SANDBOX_VERSION = SANDBOX_VERSION
		# Set up event handlers
		host.registerGameStatusHandler(self.onGameStatusChanged)
		host.registerHandler('PlayerDeath', self.onPlayerDeathCQ)
		host.registerHandler('PlayerKilled', self.onPlayerKilledCQ)
		host.registerHandler('PlayerSpawn', self.onPlayerSpawn)
		host.registerHandler('EnterVehicle', self.onEnterVehicle)
		host.registerHandler('ExitVehicle', self.onExitVehicle)
		host.registerHandler('PlayerDisconnect', self.onPlayerDisconnect)
		host.registerHandler('ChatMessage', self.onChatMessage, 1)
		host.registerHandler('RemoteCommand', self.onRemoteCommand, 1)

	def initSbx(self):
		# All allowed static objects (not including vehicles)
		self.allowed_objects = ['ladder_3m','billboard_highway_01','billboard_highway_02','billboard_highway_03','billboard_dicecream','woodencrate_1m','wooden_bridge_segment','ch_stnwalls_10m_stairs','stonefence_5m','mi_sandbags_corner_4m','steelrailing_4m','ch_wall_high_6m','ch_wall_high_6m_door','ch_wall_high_6m_win_hole','ch_wall_high_l','cabledrum_2m','parkbench2','chsalestable_box_02','ch_phone_booth','nc_birch01','me_olive03','nc_weeping01','me_flowerpot01','me_tropicalplant03','me_wildbush01','me_vines01','glassplane01','fish','teddy','me_palmtree01','toolshed_door_l','woodencrate_4m','cone_dest','sign_streetcrossing_dest','barrel_blue','pallet_mobile','barrel_yellow_mobile','woodencrate_destructible_mobile','barrel_green','fishingbox','ball','barrel_yellow','pallet','woodencrate_destructible','concrete_pillar_beam','beam_01','container_cargo_red','container_cl_green','mi_sandbags_4m','construct_metal_plate','mi_plane_stair','pedestrian_railing_12m_ends','trestle01_dest','ch_stnwalls_10m_up', 'ladder_10m_noholders', 'ladder_40m', 'concreteblock_airstrip_endstr', 'convbelt', 'mi_sandbags_circle_half', 'woodsteel_bridge_segment', 'basket_hoops', 'concretebarrier', 'pedestrian_railing_6m_ends', 'wallhigh02', 'fence_dest', 'wirefence_6m', 'brickwall', 'fence_corrugated_3x12m', 'barrier_blocks_high_8m', 'gate_boom', 'rock2m', 'cafechair', 'wpipe_10m', 'tvset', 'chsalestable_02', 'table01', 'wheelbarrow', 'trashcan_small', 'phonebooth', 'trafficlight_dest', 'trash_dumpster', 'power_station_pt_2', 'wpipe_support', 'cablepole','ch_wall_high_6m_door','supply_crate']
		# Static objects defined as "annoying" will have a double delay before subsequent spawns
		self.annoying_objects = ['convbelt','barrel_yellow_mobile','barrel_yellow']
		# Static objects that are multiplayer-compatible, that is, "_mp" can be appended to the name to
		# get a networked version of the object
		self.mp_objects = ['billboard_highway_01','billboard_highway_02','billboard_highway_03','billboard_dicecream','wooden_bridge_segment','steelrailing_4m','ch_wall_high_6m','ch_wall_high_6m_door','ch_wall_high_6m_win_hole','ch_wall_high_l','cabledrum_2m','parkbench2','chsalestable_box_02','ch_phone_booth','nc_birch01','me_olive03','nc_weeping01','me_flowerpot01','me_tropicalplant03','me_wildbush01','me_vines01','glassplane01','mi_sandbags_corner_4m','stonefence_5m','woodencrate_1m','ch_stnwalls_10m_stairs','me_palmtree01','toolshed_door_l','woodencrate_4m','table01','barrel_green','fishingbox','barrel_yellow','pallet','woodencrate_destructible','concrete_pillar_beam','beam_01','container_cargo_red','container_cl_green','mi_sandbags_4m','construct_metal_plate','mi_plane_stair','pedestrian_railing_12m_ends','trestle01_dest','ch_stnwalls_10m_up', 'concreteblock_airstrip_endstr', 'convbelt', 'mi_sandbags_circle_half', 'woodsteel_bridge_segment', 'basket_hoops', 'concretebarrier', 'pedestrian_railing_6m_ends', 'wallhigh02', 'fence_dest', 'wirefence_6m', 'brickwall', 'fence_corrugated_3x12m', 'barrier_blocks_high_8m', 'gate_boom', 'rock2m', 'cafechair', 'wpipe_10m', 'tvset', 'chsalestable_02', 'wheelbarrow', 'trashcan_small', 'phonebooth', 'trafficlight_dest', 'trash_dumpster', 'power_station_pt_2', 'wpipe_support', 'cablepole','ch_wall_high_6m_door']
		# Objects that simply don't work in multiplayer
		self.nonmp_objects = ['ladder_3m','ladder_10m_noholders','ladder_40m']
		# Objects that only admins can spawn...heheh
		self.admin_objects = ['rock2m']
		# Spawnable vehicles, will not be auto-grabbed
		self.allowed_vehicles = ['aircontroltower','faav_nb','faav_st','us_heavy_soldier','us_light_soldier','mec_heavy_soldier','mec_light_soldier','usaas_stinger','usaas_stinger_spw','ats_tow','boat_rib','civiliancar2','truck1','aav_tunguska','aav_type95','apc_btr90','apc_wz551','jeep_faav','jep_mec_paratrooper','jep_nanjing','jep_vodnik','rutnk_t90','tnk_type98','usaav_m6','usapc_lav25','usjep_hmmwv','ustnk_m1a2','ahe_ah1z','ahe_havoc','ahe_z10','air_f35b','air_j10','air_su30mkk','chthe_z8','ruair_mig29','ruair_su34','the_mi17','uav_pred','usair_f15','usair_f18','usthe_uh60']
		# Out of the previous, the ones that will be auto-grabbed if the option is enabled
		self.autograb_excempt = ['usaas_stinger','usaas_stinger_spw','ats_tow']
		# Soldier templates
		self.soldier_objects = ['us_light_soldier', 'us_heavy_soldier', 'mec_light_soldier', 'mec_heavy_soldier']
		# -------------------------
		self.object_types = []
		self.deleted_objects = []
		self.obj_freeze = {}
		self.obj_locked = {}
		self.spawned_objects = []
		self.is_spGame = 1
		self.sbxTimer = None
		self.freezerTimer = None
		self.popMessagesTimer = None
		self.groupCreatorTimer = None
		self.p_objects = {}
		self.allow_lock = True
		self.allow_grab = True
		self.allow_create = True
		self.gravity = -15
		self.spawnerTemplates = []
		self.lastObject = []
		self.allowCreateOverride = True
		self.sBoundaries = True
		self.pdata = {}
		self.preCompletedObjects = []
		self.recyclables = {}
		self.relative_grabbing = True
		self.admins = ""
		self.lastObjectCounter = 0
		self.spawnDelay = 80
		self.vehiclespawnDelay = 120
		self.dupeDelay = 1000
		self.groupLoadLimit = 200
		self.groupDupeLimit = 100
		self.groupDuper = []
		self.groupDuperObjects = []
		self.groupDuperFailures = 0
		self.flickerObjects = []
		self.netServer = "205.234.105.49"
		self.netPort = 16573
		self.groupObjectLimit = 200
		# Set up the sandbox timer
		self.timerCounter = 0
		if not self.sbxTimer:
			self.sbxTimer = bf2.Timer(self.run_sbxTimer, 0 , 1)
			self.sbxTimer.setRecurring(0.01)
		if not self.freezerTimer:
			self.freezerTimer = bf2.Timer(self.freezer, 0 , 1)
			self.freezerTimer.setRecurring(0.01)
		if not self.popMessagesTimer:
			self.popMessagesTimer = bf2.Timer(self.popMessages, 0 , 1)
			self.popMessagesTimer.setRecurring(1.0)
		if not self.groupCreatorTimer:
			self.groupCreatorTimer = bf2.Timer(self.groupCreator, 0 , 1)
			self.groupCreatorTimer.setRecurring(0.01)
		self.pMessages = {}
		self.pMessages["GRABMODE_ORIG_ALL"] = [1031105, 1]
		self.pMessages["GRABMODE_XA_MINE"] = [1031105, 2]
		self.pMessages["GRABMODE_XA_ALL"] = [1031105, 3]
		self.pMessages["GRABMODE_ORIG_MINE"] = [1031113, 1]
		self.pMessages["GROUP_INITIALIZED"] = [1031113, 2]
		self.pMessages["GROUP_ALREADY_SELECTED"] = [1031113, 3]
		self.pMessages["GROUP_DESELECTED"] = [1031121, 1]
		self.pMessages["GROUP_SELECTED"] = [1031121, 2]
		self.pMessages["GROUP_OBJECTADDED"] = [1031121, 3]
		self.pMessages["GROUP_OBJECTREMOVED"] = [1031115, 1]
		self.pMessages["ROTATION_SAVED"] = [1031115, 2]
		self.pMessages["ROTATION_CLEARED"] = [1031115, 3]
		self.pMessages["AUTOLOCK_ENABLED"] = [1031120, 1]
		self.pMessages["AUTOLOCK_DISABLED"] = [1031120, 2]
		self.pMessages["BUSY"] = [1031120, 3]
		self.pMessages["AUTOGRAB_ENABLED"] = [1031109, 1]
		self.pMessages["AUTOGRAB_DISABLED"] = [1031109, 2]
		self.pMessages["MERGE_STARTED"] = [1031109, 3]
		self.pMessages["MERGE_COMPLETED"] = [1031923, 1]
		self.pMessages["MERGE_CANCELLED"] = [1031923, 2]
		self.pMessages["GROUP_LIMITREACHED"] = [1031923, 3]
		self.pMessages["SAVEDELAY"] = [1032415, 1]
		self.pMessages["GROUP_DUPELIMITREACHED"] = [1032415, 2]
		self.pMessages["DUPEDELAY"] = [1032415, 3]
		self.pMessages["GROUP_UNGROUPED"] = [1031119, 1]
		# Debug mode? This prints a bunch of stuff to stdout so you can log it with a console or something
		self.enableDebug = 0
		self.dbg("Sandbox has been initialized.")

	def onGameStatusChanged(self, status):
		if status == bf2.GameStatus.Playing:
			self.preCompletedObjects = []
			for x in self.allowed_objects:
				for y in bf2.objectManager.getObjectsOfTemplate(x):
					self.preCompletedObjects.append(y)
			for x in bf2.objectManager.getObjectsOfType('dice.hfe.world.ObjectTemplate.PlayerControlObject'):
				self.preCompletedObjects.append(x)
			g_controlPoints = bf2.objectManager.getObjectsOfType('dice.hfe.world.ObjectTemplate.ControlPoint')
			self.con_save_cp = 1
			try:
				for obj in g_controlPoints:
					if obj.cp_getParam('team') == 2:
						self.con_save_cp = obj.getTemplateProperty('ControlPointId')
					elif obj.cp_getParam('team') == 0:
						obj.cp_setParam('team', 2)
			except:
				pass
			self.p_objects = {}
			self.pdata = {}
			self.object_types = []
			self.deleted_objects = []
			self.obj_freeze = {}
			self.obj_locked = {}
			self.gravity = -15
			self.spawnerTemplates = []
			self.spawned_objects = []
			self.ftpUpload = False
			self.ftpOptions = {
				'server': '',
				'username': '',
				'password': '',
				'target': '',
				'dlLocation': '',
			}
			host.rcon_invoke("physics.gravity " + str(self.gravity))
			self.groupDuper = []
			self.groupDuperObjects = []
			self.flickerObjects = []
			self.preCompletedObjects = []
			self.lastObject = []
			self.recyclables = {}
			try:
				self.loadSettings()
			except:
				ExceptionOutput()
			if self.sBoundaries:
				self.enableBoundaries()
			else:
				self.disableBoundaries()
			print self.admins
			self.dbg(self.admins)
			self.dbg("Game status cleared and Sandbox reloaded.")

	#######################################
	#######################################
	#        Configuration Loader         #
	#######################################
	#######################################

	def loadSettings(self):
		config = ConfigParser.ConfigParser()
		config.read(host.sgl_getModDirectory() + "/settings/sandbox.ini")

		# [admins]
		self.admins = config.get("users","admins")

		# [antispam]
		self.spawnDelay = int(config.get("antispam","spawndelay"))
		self.vehiclespawnDelay = int(config.get("antispam","vehiclespawndelay"))

		# [grouping]
		self.dupeDelay = int(config.get("grouping","dupedelay"))
		self.groupObjectLimit = int(config.get("grouping","objectlimit"))
		self.groupDupeLimit = int(config.get("grouping","duplicatelimit"))
		self.groupLoadLimit = int(config.get("grouping","loadlimit"))

		# [hitpoints]
		host.rcon_invoke("objecttemplate.active mec_light_soldier")
		host.rcon_invoke("objecttemplate.armor.maxhitpoints " + str(config.get("hitpoints","zombie")))
		host.rcon_invoke("objecttemplate.armor.hitpoints " + str(config.get("hitpoints","zombie")))
		host.rcon_invoke("objecttemplate.active mec_3p_soldier")
		host.rcon_invoke("objecttemplate.armor.maxhitpoints " + str(config.get("hitpoints","zombie")))
		host.rcon_invoke("objecttemplate.armor.hitpoints " + str(config.get("hitpoints","zombie")))
		host.rcon_invoke("objecttemplate.active us_light_soldier")
		host.rcon_invoke("objecttemplate.armor.maxhitpoints " + str(config.get("hitpoints","construction")))
		host.rcon_invoke("objecttemplate.armor.hitpoints " + str(config.get("hitpoints","construction")))
		host.rcon_invoke("objecttemplate.active us_3p_soldier")
		host.rcon_invoke("objecttemplate.armor.maxhitpoints " + str(config.get("hitpoints","construction")))
		host.rcon_invoke("objecttemplate.armor.hitpoints " + str(config.get("hitpoints","construction")))
		host.rcon_invoke("objecttemplate.active us_heavy_soldier")
		host.rcon_invoke("objecttemplate.armor.maxhitpoints " + str(config.get("hitpoints","assaultmedic")))
		host.rcon_invoke("objecttemplate.armor.hitpoints " + str(config.get("hitpoints","assaultmedic")))
		host.rcon_invoke("objecttemplate.active ch_heavy_soldier")
		host.rcon_invoke("objecttemplate.armor.maxhitpoints " + str(config.get("hitpoints","sniper")))
		host.rcon_invoke("objecttemplate.armor.hitpoints " + str(config.get("hitpoints","sniper")))
		host.rcon_invoke("objecttemplate.active mec_heavy_soldier")
		host.rcon_invoke("objecttemplate.armor.maxhitpoints " + str(config.get("hitpoints","explosives")))
		host.rcon_invoke("objecttemplate.armor.hitpoints " + str(config.get("hitpoints","explosives")))

		# [engine]
		self.gravity = int(config.get("engine","gravity"))
		host.rcon_invoke("physics.gravity " + str(self.gravity))

		# [ftp]
		self.ftpUpload = bool(config.get("ftp","enable"))
		self.ftpOptions['server'] = config.get("ftp","server")
		self.ftpOptions['username'] = config.get("ftp","username")
		self.ftpOptions['password'] = config.get("ftp","password")
		self.ftpOptions['target'] = config.get("ftp","directory")
		self.ftpOptions['dlLocation'] = config.get("ftp","address")

	def reloadAdmins(self):
		config = ConfigParser.ConfigParser()
		config.read(host.sgl_getModDirectory() + "/settings/sandbox.ini")
		self.admins = config.get("users","admins")


	#######################################
	#######################################
	#        Chat message handler         #
	#######################################
	#######################################

	def onRemoteCommand(self, playerId, cmd):
		from math import pi
		if playerId == -1: #fix for local servers
			playerId = 255
		self.dbg("Command received: [" + str(cmd) + "] from [" + str(playerId) + "]")
		p = bf2.PlayerManager.Player(playerId)
		pD = self.pdata[p.getName()]
		pPos = p.getDefaultVehicle().getPosition()
		pRot = p.getDefaultVehicle().getRotation()
		origCmd = cmd
		spacepos = origCmd.find(' ')
		if spacepos == -1: spacepos=len(cmd)
		cmd = origCmd[0:spacepos]
		cmdVar = origCmd[spacepos + 1:]
		if cmd == "btnSpawn":
			if len(pD.selectedGroup) >= self.groupObjectLimit and not self.is_spGame:
				self.sendMsg(p, "GROUP_LIMITREACHED")
				return 0
			if not self.allow_create and not (self.checkAdmin(p.getName()) and self.allowCreateOverride):
				return 0
			if cmdVar and (not self.allowed_objects.count(cmdVar.lower()) and not self.allowed_vehicles.count(cmdVar.lower())):
				return 0
			if pD.spawnDelay > 0:
				return 0
			if self.nonmp_objects.count(cmdVar.lower()) and not self.is_spGame:
				return 0
			if self.admin_objects.count(cmdVar.lower()) and not self.checkAdmin(p.getName()):
				return 0
			if cmdVar == "aircontroltower" and str(host.sgl_getMapName()).lower() == "thesandbox":
				return 0
			if self.mp_objects.count(cmdVar.lower()) and not self.is_spGame:
				cmdVar += "_mp"
			if self.groupDuper:
				fail = False
				for x in self.groupDuperObjects:
					if x[0] == cmdVar:
						fail = True
				if fail:
					self.sendMsg(p,"BUSY")
					return 0
			pCam = self.getSoldierCam(p)
			pCamRot = pCam.getRotation()
			sPos = self.getVectoredPosition(pPos,pD.tracerDistance,pRot,pCamRot)
			sRot = pRot
			self.dbg("new position: " + str(sPos[0]) + "," + str(sPos[1]) + "," + str(sPos[2]))
			self.dbg("new rotation: " + str(sRot[0]) + "," + str(sRot[1]) + "," + str(sRot[2]))
			if cmdVar:
				pD.lastTemplate = str(cmdVar)
			else:
				if not pD.lastTemplate:
					return
				if pD.lastRotation:
					sRot = pD.lastRotation
			cmdVar = pD.lastTemplate
			try:
				self.createSpawnerObject(str(cmdVar), sPos, sRot, p.index, True)
			except:
				ExceptionOutput()
			if pD.objectTypes.count(str(cmdVar)) < 1:
				pD.objectTypes.append(str(cmdVar))
			if self.object_types.count(str(cmdVar)) < 1:
				self.object_types.append(str(cmdVar))
			pD.lastRotation = sRot
			if not self.is_spGame:
				if self.annoying_objects.count(str(cmdVar)) or self.allowed_vehicles.count(str(cmdVar)):
					pD.spawnDelay = self.vehiclespawnDelay
				else:
					pD.spawnDelay = self.spawnDelay
			self.dbg("Created spawner")
			#print time.strftime("%H:%M:%S") + " - [" + p.getName() + "] - [" + str(cmdVar) + "]"

		if cmd == "btnGrabAlt":
			if self.is_spGame:
				if pD.preGrabMode == 0:
					pD.preGrabMode = 1
				else:
					o = pD.preGrab[0]
					if hasattr(o,"oId"):
						host.rcon_invoke("Object.active id" + str(o.oId))
						host.rcon_invoke("Object.isVisible 1")
					pD.preGrab[0].setPosition(pD.preGrab[3])
					pD.preGrabMode = 0
					pD.preGrab = []

		if cmd == "btnGrab":
			if pD.spawnDelay > 40:
				return 0
			if not self.allow_grab and not self.checkAdmin(p.getName()):
				return 0
			if pD.preGrabMode == 1 and self.is_spGame:
				o = pD.preGrab[0]
				if hasattr(o,"oId"):
					host.rcon_invoke("Object.active id" + str(o.oId))
					host.rcon_invoke("Object.isVisible 1")
				pD.preGrab[0].setPosition(pD.preGrab[3])
				pD.preGrabMode = 0
				pD.preGrab = []
			shortest = [0,pD.tracerDistance + 20.0]
			pCam = self.getSoldierCam(p)
			pCamRot = pCam.getRotation()
			sPos = self.getVectoredPosition(pPos,pD.tracerDistance,pRot,pCamRot)
			sRot = pRot
			if pD.grabMode == 2 or pD.grabMode == 4:
				for v in pD.objectTypes:
					for q in bf2.objectManager.getObjectsOfTemplate(v):
						if q.isValid():
							qDist = self.getVectorDistance(sPos,q.getPosition())
							if qDist < shortest[1]:
								shortest = [q,qDist]
			else:
				for q in pD.objects:
					if q.isValid():
						qDist = self.getVectorDistance(sPos,q.getPosition())
						if qDist < shortest[1]:
							shortest = [q,qDist]
			if shortest[0]:
				if not self.is_spGame and self.obj_locked.has_key(shortest[0]) and not self.checkAdmin(p.getName()):
					if self.obj_locked[shortest[0]] != p.getName(): return 0
				distance = self.getVectorDistance(pPos,shortest[0].getPosition())
				frozen = 0
				if self.obj_freeze.has_key(shortest[0]):
					frozen = 1
					self.obj_freeze.pop(shortest[0])
				if pD.grabMode <= 2:
					pD.grabbed = [shortest[0],distance,1,shortest[0].getPosition(),shortest[0].getRotation()]
				else:
					shp = shortest[0].getPosition()
					sPos = self.getVectoredPosition(pPos,distance,pRot,pCamRot)
					grabOffset = ((sPos[0] - shp[0]),(sPos[1] - shp[1]),(sPos[2] - shp[2]))
					pD.grabbed = [shortest[0],distance,1,shortest[0].getPosition(),shortest[0].getRotation(),grabOffset]
				pD.lastAction.append([1,shortest[0],shortest[0].getPosition(),shortest[0].getRotation(),frozen])
				# 1 = grab [object, original position, original rotation, original frozen]
				self.dbg("Grabbed closest object")
				pD.spawnDelay = self.spawnDelay

		if cmd == "btnGrabMode":
			if pD.grabMode == 1:
				pD.grabMode = 2
				self.sendMsg(p, "GRABMODE_ORIG_ALL")
			elif pD.grabMode == 2:
				pD.grabMode = 3
				pD.grabbed = []
				self.sendMsg(p, "GRABMODE_XA_MINE")
			elif pD.grabMode == 3:
				pD.grabMode = 4
				self.sendMsg(p, "GRABMODE_XA_ALL")
			elif pD.grabMode == 4:
				pD.grabMode = 1
				pD.grabbed = []
				self.sendMsg(p, "GRABMODE_ORIG_MINE")

		if cmd == "btnStartGroup":
			if not pD.groupCreate:
				pD.groupCreate = True
				self.sendMsg(p, "GROUP_INITIALIZED")
			else:
				self.sendMsg(p, "GROUP_ALREADY_SELECTED")

		if cmd == "btnEndGroup":
			try:
				if pD.groupCreate:
					pD.groupCreate = False
					pD.groups.append(pD.selectedGroup)
					pD.selectedGroup = []
					if not cmdVar:
						self.sendMsg(p, "GROUP_DESELECTED")
			except:
				ExceptionOutput()

		if cmd == "btnSelectGroup":
			if pD.grabbedGroup:
				pD.selectedGroup = []
				for x in pD.grabbedGroup:
					pD.selectedGroup.append(x[0])
				pD.groupCreate = True
				if not cmdVar:
					self.sendMsg(p, "GROUP_SELECTED")

		if cmd == "btnAddToGroup":
			if len(pD.selectedGroup) >= self.groupObjectLimit and not self.is_spGame:
				self.sendMsg(p, "GROUP_LIMITREACHED")
				return
			noGrab = False
			if not pD.grabbed:
				self.onRemoteCommand(playerId, "btnGrab")
				noGrab = True
			if pD.grabbed and pD.groupCreate:
				if not pD.selectedGroup.count(pD.grabbed[0]):
					pD.selectedGroup.append(pD.grabbed[0])
					self.sendMsg(p, "GROUP_OBJECTADDED")
			if noGrab:
				self.onRemoteCommand(playerId, "btnDrop")
				self.flickerObjects.append([pD.grabbed[0],pD.grabbed[0].getPosition(),10])

		if cmd == "btnRemFromGroup":
			if pD.grabbed and pD.groupCreate:
				if pD.selectedGroup.count(pD.grabbed[0]):
					del pD.selectedGroup[pD.selectedGroup.index(pD.grabbed[0])]
					self.sendMsg(p, "GROUP_OBJECTREMOVED")

		if cmd == "btnGrabGroup":
			if not pD.grabbed:
				self.onRemoteCommand(playerId, "btnGrab")
			if pD.grabbed:
				anchor = pD.grabbed[0]
				selGroup = False
				for y in pD.groups:
					if y.count(anchor) and not pD.lastGroup.count(y):
						selGroup = y
				if not selGroup:
					pD.lastGroup = []
					for y in pD.groups:
						if y.count(anchor):
							selGroup = y
				if selGroup:
					pD.grabbedGroupAnchor = pD.grabbed
					pD.grabbed = []
					if pD.lastGroup:
						pD.lastGroup.append(selGroup)
						pD.grabbedGroup = []
					else:
						pD.lastGroup = []
						pD.lastGroup.append(selGroup)
					anchorPos = anchor.getPosition()
					#anchorPos = pD.grabbedGroupAnchor[3]
					count = 0
					for x in selGroup:
						try:
							oPos = x.getPosition()
						except:
							pass
						else:
							relPos = ((oPos[0] - anchorPos[0]),(oPos[1] - anchorPos[1]),(oPos[2] - anchorPos[2]))
							oRot = x.getRotation()
							if self.obj_freeze.has_key(x):
								self.obj_freeze.pop(x)
							pD.grabbedGroup.append([x,relPos,oRot])
							count += 1
					self.dbg("Group grabbed, counted " + str(count) + " objects")
					#self.sendMsg(0,"Group grabbed (" + p.getName() + " - " + str(count) + ")")
					pD.lastAction.append([4,pD.grabbedGroupAnchor[0],pD.grabbedGroupAnchor[0].getPosition(),pD.grabbedGroupAnchor[0].getRotation(),pD.grabbedGroup])
					# 4 = groupgrab [anchor, anchor position, anchor rotation, original grabbedgroup]
					self.dbg("Anchor rotation is...")
					self.dbg(str(pD.grabbedGroupAnchor[0].getRotation()))

		if cmd == "btnDuplicateGroup":
			if self.groupDuper:
				self.sendMsg(p,"BUSY")
				return 0
			if pD.dupeDelay and not self.is_spGame:
				self.sendMsg(p,"DUPEDELAY")
				return 0
			if pD.grabbedGroup:
				self.onRemoteCommand(playerId, "btnSelectGroup nr")
			if pD.selectedGroup:
				if len(pD.selectedGroup) > self.groupDupeLimit and not self.is_spGame:
					self.sendMsg(p, "GROUP_DUPELIMITREACHED")
					self.onRemoteCommand(playerId, "btnEndGroup nr")
					return 0
				self.onRemoteCommand(playerId, "btnDrop")
				anchor = pD.selectedGroup[0]
				anchorPos = anchor.getPosition()
				self.groupDuperObjects.append([anchor.templateName,(0.0,0.0,0.0),anchor.getRotation()])
				for x in pD.selectedGroup:
					if x != anchor:
						oPos = x.getPosition()
						# [object name, object position, object rotation]
						relPos = ((oPos[0] - anchorPos[0]),(oPos[1] - anchorPos[1]),(oPos[2] - anchorPos[2]))
						self.groupDuperObjects.append([x.templateName,relPos,x.getRotation()])
				# [anchor, user, stage, index]
				pD.selectedGroup = []
				pD.groupCreate = True
				self.groupDuper = [self.groupDuperObjects[0],p.index,1,0]
				pD.dupeDelay = self.dupeDelay

		if cmd == "btnMergeGroup":
			if not pD.startMerge:
				if pD.grabbedGroup:
					self.onRemoteCommand(playerId, "btnSelectGroup nr")
				if pD.selectedGroup:
					pD.startMerge = pD.selectedGroup
					self.sendMsg(p, "MERGE_STARTED")
					self.onRemoteCommand(playerId, "btnEndGroup nr")
					self.onRemoteCommand(playerId, "btnDrop")
			else:
				if pD.grabbedGroup:
					self.onRemoteCommand(playerId, "btnSelectGroup nr")
				if pD.selectedGroup and (pD.selectedGroup != pD.startMerge):
					if (len(pD.selectedGroup) + len(pD.startMerge)) >= self.groupObjectLimit:
						self.sendMsg(p, "GROUP_LIMITREACHED")
						self.sendMsg(p, "MERGE_CANCELLED")
						pD.startMerge = []
						self.onRemoteCommand(playerId, "btnEndGroup nr")
						return
					for x in pD.startMerge:
						pD.selectedGroup.append(x)
					pD.startMerge = []
					self.sendMsg(p, "MERGE_COMPLETED")
					self.onRemoteCommand(playerId, "btnEndGroup nr")
					self.onRemoteCommand(playerId, "btnDrop")
					self.onRemoteCommand(playerId, "btnGrabGroup")
				else:
					self.sendMsg(p, "MERGE_CANCELLED")
					pD.startMerge = []
					self.onRemoteCommand(playerId, "btnEndGroup nr")

		if cmd == "btnUnGroup":
			if pD.grabbedGroup:
				self.onRemoteCommand(playerId, "btnSelectGroup nr")
				pD.dropGroup()
			if pD.selectedGroup:
				for g in pD.groups:
					if g == pD.selectedGroup:
						pD.groups.remove(g)
						pD.selectedGroup = []
						self.sendMsg(p, "GROUP_UNGROUPED")

		if cmd == "btnRotate":
			v = cmdVar.split(" ")
			if pD.grabbed:
				if v[1] == "1": rFactor = pD.rotSnap
				else: rFactor = -pD.rotSnap
				obj = pD.grabbed[0]
				if obj and obj.isValid():
					objRot = obj.getRotation()
					if v[0] == "A":
						addRot = round(objRot[0] + rFactor)
						newRot = (addRot, objRot[1], objRot[2])
					if v[0] == "P":
						addRot = round(objRot[1] + rFactor)
						if addRot >= 89.0:
							addRot = 89.9
						elif addRot <= -89.0:
							addRot = -89.9
						newRot = (objRot[0], addRot, objRot[2])
					if v[0] == "R":
						addRot = round(objRot[2] + rFactor)
						newRot = (objRot[0], objRot[1], addRot)
					self.dbg(newRot)
					pD.grabbed[4] = newRot
					obj.setRotation(newRot)
					pD.lastRotation = newRot
					#self.dbg(obj.getRotation())
					obj.setRotation(newRot)
					#updatedRot = obj.getRotation()
					self.dbg(updatedRot)
			elif pD.grabbedGroupAnchor:
				try:
					if v[1] == "1": rFactor = pD.rotSnap
					else: rFactor = -pD.rotSnap
					obj = pD.grabbedGroupAnchor[0]
					if obj and obj.isValid():
						objRot = obj.getRotation()
						if v[0] == "A":
							currKey = 0
							for x in pD.grabbedGroup:
								pos = x[0].getPosition()
								rot = x[0].getRotation()
								newPos = self.get2dRotation(x[1][0],x[1][2],-rFactor) # Use x and z coordinates
								pD.grabbedGroup[currKey][1] = (newPos[0],x[1][1],newPos[1]) # Set the new position of the object
								pD.grabbedGroup[currKey][2] = ((rot[0] + rFactor, rot[1], rot[2])) # Set the new rotation of the object
								currKey += 1
							addRot = round(objRot[0] + rFactor)
							newRot = (addRot, objRot[1], objRot[2])
							pD.grabbedGroupAnchor[4] = newRot
				except:
					ExceptionOutput()

		if cmd == "btnResetRotation":
			if pD.grabbed:
				obj = pD.grabbed[0]
				if obj and obj.isValid():
					if not pD.savedRotation:
						pD.grabbed[4] = (0.0,0.0,0.0)
						obj.setRotation((0.0,0.0,0.0))
						pD.lastRotation = (0.0,0.0,0.0)
					else:
						pD.grabbed[4] = pD.savedRotation
						obj.setRotation(pD.savedRotation)
						pD.lastRotation = pD.savedRotation

		if cmd == "btnSaveRotation":
			if not pD.savedRotation:
				if pD.grabbed:
					obj = pD.grabbed[0]
					if obj and obj.isValid():
						pD.savedRotation = obj.getRotation()
						self.sendMsg("ROTATION_SAVED")
			else:
				pD.savedRotation = ()
				self.sendMsg(p, "ROTATION_CLEARED")

		if cmd == "btnDrop":
			if pD.grabbed:
				pD.lastAction.append([2,pD.grabbed[0],pD.grabbed[1],0])
				# 2 = drop [object, distance]
				pD.grabbed = []
			if pD.grabbedGroup:
				pD.dropGroup()

		if cmd == "btnFreeze":
			if pD.grabbed:
				if pD.grabbed[0].isValid():
					self.obj_freeze[pD.grabbed[0]] = [pD.grabbed[0].getPosition(),pD.grabbed[0].getRotation()]
					pD.lastAction.append([2,pD.grabbed[0],pD.grabbed[1],1])
					pD.grabbed = []

		if cmd == "btnLock":
			if pD.grabbed and self.allow_lock:
				if pD.grabbed[0].isValid():
					self.obj_locked[pD.grabbed[0]] = p.getName()

		if cmd == "btnUnLock":
			if pD.grabbed:
				if pD.grabbed[0].isValid():
					try:
						del self.obj_locked[pD.grabbed[0]]
					except:
						pass #do nothing

		if cmd == "btnDelete":
			if pD.grabbed:
				if pD.grabbed[0].isValid():
					o = pD.grabbed[0]
					pD.grabbed = []
					idx = 0
					for g in pD.groups:
						try:
							while g.count(o):
								pD.groups[idx].remove(o)
						except:
							ExceptionOutput()
						idx += 1
					idx = 0
					for y in pD.grabbedGroup:
						try:
							if y[0] == o:
								pD.grabbedGroup.remove(y)
						except:
							ExceptionOutput()
					try:
						while pD.selectedGroup.count(o):
							pD.selectedGroup.remove(o)
					except:
						ExceptionOutput()
					self.deleteObject(o)
			elif pD.grabbedGroupAnchor:
				for x in pD.grabbedGroup:
					self.deleteObject(x[0])
				pD.dropGroup()

		if cmd == "btnPush":
			if pD.grabbed:
				if cmdVar == "1":
					pD.grabbed[1] += 0.5
				else:
					pD.grabbed[1] += 1.0
			elif pD.grabbedGroupAnchor:
				if cmdVar == "1":
					pD.grabbedGroupAnchor[1] += 0.5
				else:
					pD.grabbedGroupAnchor[1] += 1.0

		if cmd == "btnPull":
			if pD.grabbed:
				if cmdVar == "1":
					pD.grabbed[1] -= 0.5
				else:
					pD.grabbed[1] -= 1.0
			elif pD.grabbedGroupAnchor:
				if cmdVar == "1":
					pD.grabbedGroupAnchor[1] -= 0.5
				else:
					pD.grabbedGroupAnchor[1] -= 1.0

		if cmd == "btnUndo":
			#if pD.lastAction and self.is_spGame:
			if pD.lastAction:
				a = pD.lastAction.pop()
				if a[0] == 1: #grab [object, original position, original rotation, original frozen]
					try:
						self.dbg("Grab undoing")
						o = a[1]
						if pD.grabbed[0] == o:
							pD.grabbed = []
						o.setPosition(a[2])
						o.setRotation(a[3])
						if a[4]:
							self.obj_freeze[o] = [a[2],a[3]]
							self.dbg("Refrozen")
					except:
						ExceptionOutput()
				if a[0] == 2: # 2 = drop [object, distance]
					try:
						self.dbg("Drop undoing")
						o = a[1]
						pD.grabbed = []
						pD.grabbed = [o,a[2],1,o.getPosition(),o.getRotation(),(0.0,0.0,0.0)]
					except:
						ExceptionOutput()
				if a[0] == 3: # 3 = delete [object name, original position, original rotation]
					self.dbg("Delete undoing")
					#o = a[1]
					#o.setPosition(a[2])
					#self.deleted_objects.remove(o)
					#pD.numSpawned[0] += 1
					#self.onRemoteCommand(playerId,"btnUndo") #Undo once more so we re-grab the deleted object
				if a[0] == 4: # 4 = groupgrab [anchor, anchor position, anchor rotation, original group]
					self.dbg("Group grab undoing")
					pD.dropGroup()
					try:
						a[1].setPosition(a[2])
						self.dbg("Anchor rotation is...")
						self.dbg(str(a[3]))
						a[1].setRotation(a[3])
						self.dbg(str(a[1].getRotation()))
						aPos = a[2]
						for x in a[4]:
							x[0].setPosition((aPos[0] + x[1][0],aPos[1] + x[1][1],aPos[2] + x[1][2]))
							x[0].setRotation(x[2])
					except:
						ExceptionOutput()
				self.dbg("Undone.")

		if cmd == "btnTracerToggle":
			children = p.getPrimaryWeapon().getChildren()
			for child in children:
				if child.templateName == "glowshit":
					pos = child.getPosition()
					if pD.tracerOn:
						child.setPosition((pos[0],pos[1] - 100.0,pos[2]))
						pD.tracerOn = 0
					else:
						child.setPosition((pos[0],pos[1] + 100.0,pos[2]))
						pD.tracerOn = 1

		if cmd == "btnAutoLock":
			if pD.autoLock:
				pD.autoLock = 0
				self.sendMsg(p,"AUTOLOCK_DISABLED")
			else:
				pD.autoLock = 1
				self.sendMsg(p,"AUTOLOCK_ENABLED")

		if cmd == "btnAutoGrabToggle":
			if pD.autoGrab:
				pD.autoGrab = 0
				self.sendMsg(p,"AUTOGRAB_DISABLED")
			else:
				pD.autoGrab = 1
				self.sendMsg(p,"AUTOGRAB_ENABLED")

		if cmd == "btnTracerDist":
			if cmdVar == "push":
				pD.tracerDistance += 1
			else:
				pD.tracerDistance -= 1
			self.sendMsg(p, "Tracer distance set to " + str(pD.tracerDistance))
			if self.is_spGame:
				children = p.getPrimaryWeapon().getChildren()
				for child in children:
					if child.templateName == "glowshit":
						pPos = p.getPrimaryWeapon().getPosition()
						pCam = self.getSoldierCam(p)
						pCamRot = pCam.getRotation()
						sPos = self.getVectoredPosition(pPos,pD.tracerDistance,pRot,pCamRot)
						child.setPosition(sPos)

		if cmd == "btnRotSnap": # Steps through rotation snaps 1,5,10,15,etc.
			if cmdVar == "inc":
				if pD.rotSnap > 4:
					pD.rotSnap += 5
				else:
					pD.rotSnap = 5
			else:
				if pD.rotSnap > 5:
					pD.rotSnap -= 5
				elif pD.rotSnap <= 5 and pD.rotSnap > 1:
					pD.rotSnap = 1
			self.sendMsg(p, "Rotation factor set to " + str(pD.rotSnap) + "deg")

		if cmd == "btnSnap":
			if cmdVar == "inc":
				pD.gridSnap += 0.25
			else:
				pD.gridSnap -= 0.25
			if pD.gridSnap < 0.25: pD.gridSnap = 0.0
			if pD.gridSnap > 0:
				self.sendMsg(p, "Grid snap set to " + str(pD.gridSnap))
			else:
				self.sendMsg(p, "Grid snap disabled")

		if self.checkAdmin(p.getName()) or self.is_spGame == 1:

			if cmd == "btnGravity" and cmdVar:
				try:
					gravFactor = int(cmdVar)
					self.gravity += gravFactor
					self.sendMsg(0,"Gravity set to " + str(self.gravity))
					host.rcon_invoke("physics.gravity " + str(self.gravity))
				except:
					ExceptionOutput()

			if cmd == "btnAdGrab" and cmdVar:
				if cmdVar == "1":
					self.allow_grab = True
					self.sendMsg(0,"Object grab enabled")
				else:
					self.allow_grab = False
					self.sendMsg(0,"Object grab disabled")

			if cmd == "btnAdLock" and cmdVar:
				if cmdVar == "1":
					self.allow_lock = True
					self.sendMsg(0,"Object lock enabled")
				else:
					self.allow_lock = False
					self.sendMsg(0,"Object lock disabled")

			if cmd == "btnAdCreate" and cmdVar:
				if cmdVar == "1":
					self.allow_create = True
					self.sendMsg(0,"Object create enabled")
				else:
					self.allow_create = False
					self.sendMsg(0,"Object create disabled")

			if cmd == "delAll" and cmdVar:
				try:
					if self.pdata.has_key(cmdVar):
						plD = self.pdata[cmdVar]
						plD.grabbed = []
						plD.lastAction = []
						for o in plD.objects:
							self.deleteObject(o)
						self.sendMsg(0, str(cmdVar) + " - All objects deleted")
				except:
					ExceptionOutput()
			elif cmd == "delAll":
				pD.grabbed = []
				for o in pD.objects:
					self.deleteObject(o)

			if cmd == "delAllN" and cmdVar:
				try:
					plN = bf2.PlayerManager.Player(int(cmdVar))
					plD = self.pdata[p.getName()]
					if plD:
						plD.grabbed = []
						for o in plD.objects:
							self.deleteObject(o)
						self.sendMsg(0, str(cmdVar) + " - All objects deleted")
				except:
					ExceptionOutput()

		if cmd == "btnNone":
			self.dbg("btnNone")


	#######################################
	#######################################
	#         Sandbox core timers         #
	#######################################
	#######################################

	def run_sbxTimer(self, data):
		try:
			self.timerCounter += 1
			if self.timerCounter >= 5:
				self.timerCounter = 0
			for p in bf2.playerManager.getPlayers():
				if not self.p_objects.has_key(p.getName()):
					self.p_objects[p.getName()] = p
				if not self.pdata.has_key(p.getName()):
					self.pdata[p.getName()] = pData()
			if self.lastObject and self.lastObjectCounter <= 5:
				self.lastObjectCounter += 1
				oName = self.lastObject[0]
				user = self.lastObject[1]
				oPos = self.lastObject[2]
				oRot = self.lastObject[3]
				p = bf2.PlayerManager.Player(user)
				pD = self.pdata[p.getName()]
				if type(oPos) == tuple and type(oRot) == tuple and p.isAlive():
					objectsOfTemplate = bf2.objectManager.getObjectsOfTemplate(oName)
					o = None
					o = objectsOfTemplate[-1]
					if o.isValid() and not self.preCompletedObjects.count(o):
						pass
					else:
						o = None
					if o:
						print o.getPosition()
						self.preCompletedObjects.append(o)
						#if self.is_spGame:
						#objs = host.rcon_invoke("object.listObjectsofTemplate " + str(oName))
						#objs = objs.split()
						#o.oId = objs[-7]
						#if o.oId:
						#	self.dbg("Got object ID, " + str(o.oId))
						pD.objects.append(o)
						if pD.autoGrab and ((not self.allowed_vehicles.count(oName)) or (self.autograb_excempt.count(oName))):
						#if pD.autoGrab:
							if pD.grabMode <= 2:
								pD.grabbed = [o,pD.tracerDistance,1,o.getPosition(),o.getRotation()]
							else:
								pPos = p.getDefaultVehicle().getPosition()
								pRot = p.getDefaultVehicle().getRotation()
								pCam = self.getSoldierCam(p)
								pCamRot = pCam.getRotation()
								sPos = self.getVectoredPosition(pPos,pD.tracerDistance,pRot,pCamRot)
								shp = o.getPosition()
								grabOffset = ((sPos[0] - shp[0]),(sPos[1] - shp[1]),(sPos[2] - shp[2]))
								pD.grabbed = [o,pD.tracerDistance,1,o.getPosition(),o.getRotation(),grabOffset]
						if pD.groupCreate:
							pD.selectedGroup.append(o)
							self.dbg("Added object to group")
						if pD.autoLock and self.allow_lock:
							self.obj_locked[o] = p.getName()
							self.dbg("Object auto-locked.")
						self.lastObject = []
						self.lastObjectCounter = 0
		except:
			ExceptionOutput()

		for x in bf2.playerManager.getPlayers():
			try:
				p = bf2.PlayerManager.Player(x.index)
				pD = self.pdata[p.getName()]
				if pD.spawnDelay > 0:
					pD.spawnDelay -= 1
				if p.isAlive():
					pPos = p.getDefaultVehicle().getPosition()
					pRot = p.getDefaultVehicle().getRotation()
					#### Get all grabbed objects and move them where they need to be
					if pD.grabbed:
						if pD.grabbed[0].isValid():
							pCam = self.getSoldierCam(p)
							pCamRot = pCam.getRotation()
							sPos = self.getVectoredPosition(pPos,pD.grabbed[1],pRot,pCamRot)
							if pD.gridSnap:
								sPos = (math.ceil(sPos[0] / pD.gridSnap) * pD.gridSnap,math.ceil(sPos[1] / pD.gridSnap) * pD.gridSnap,math.ceil(sPos[2] / pD.gridSnap) * pD.gridSnap)
							if pD.grabbed[0].getPosition() != sPos:
								if pD.grabMode <= 2:
									pD.grabbed[0].setPosition(sPos)
								else:
									pD.grabbed[0].setPosition(((sPos[0] - pD.grabbed[5][0]),(sPos[1] - pD.grabbed[5][1]),(sPos[2] - pD.grabbed[5][2])))
							pD.grabbed[0].setRotation(pD.grabbed[4])
							try:
								if hasattr(pD.grabbed[0],"setDamage"):
									if pD.grabbed[0].getDamage() > 0:
										pD.grabbed[0].setDamage(10000)
							except:
								pass
						else:
							pD.grabbed = []
					#### Get all groups and move them where they need to be
					elif pD.grabbedGroup:
						if pD.grabbedGroupAnchor[0].isValid():
							pCam = self.getSoldierCam(p)
							pCamRot = pCam.getRotation()
							sPos = self.getVectoredPosition(pPos,pD.grabbedGroupAnchor[1],pRot,pCamRot)
							if pD.gridSnap:
								sPos = (math.ceil(sPos[0] / pD.gridSnap) * pD.gridSnap,math.ceil(sPos[1] / pD.gridSnap) * pD.gridSnap,math.ceil(sPos[2] / pD.gridSnap) * pD.gridSnap)
							if pD.grabbedGroupAnchor[0].getPosition() != sPos:
								if pD.grabMode <= 2:
									pD.grabbedGroupAnchor[0].setPosition(sPos)
								else:
									pD.grabbedGroupAnchor[0].setPosition(((sPos[0] - pD.grabbedGroupAnchor[5][0]),(sPos[1] - pD.grabbedGroupAnchor[5][1]),(sPos[2] - pD.grabbedGroupAnchor[5][2])))
								#self.dbg("new position: " + str(sPos[0]) + "," + str(sPos[1]) + "," + str(sPos[2]))
								aPos = pD.grabbedGroupAnchor[0].getPosition()
								for x in pD.grabbedGroup:
									x[0].setPosition((aPos[0] + x[1][0],aPos[1] + x[1][1],aPos[2] + x[1][2]))
									x[0].setRotation(x[2])
							pD.grabbedGroupAnchor[0].setRotation(pD.grabbedGroupAnchor[4])
						else:
							pD.dropGroup()
					if pD.preGrabMode == 1 and self.timerCounter == 0:
						pCam = self.getSoldierCam(p)
						shortest = []
						shortest = [0,pD.tracerDistance + 10.0]
						pCamRot = pCam.getRotation()
						sPos = self.getVectoredPosition(pPos,pD.tracerDistance,pRot,pCamRot)
						sRot = pRot
						for v in pD.objectTypes:
							if self.allowed_objects.count(v):
								for q in bf2.objectManager.getObjectsOfTemplate(v):
									if q.isValid():
										qDist = self.getVectorDistance(sPos,q.getPosition())
										if qDist < shortest[1]:
											shortest = [q,qDist]
						if pD.preGrab:
							if shortest[1] > pD.preGrab[1]:
								shortest[0] = pD.preGrab[0]
								shortest[1] = pD.preGrab[1]
						else: pD.preGrab = [shortest[0],shortest[1],1,shortest[0].getPosition(),pCam.getRotation()]
						if pD.preGrab[0] != shortest[0]:
							# Object, current flash thing, object position, player rotation
							o = pD.preGrab[0]
							if o.isValid():
								if hasattr(o,"oId"):
									host.rcon_invoke("Object.active id" + str(o.oId))
									host.rcon_invoke("Object.isVisible 1")
								else: o.setPosition(pD.preGrab[3])
								pD.preGrab[1] = 1
							pD.preGrab = [shortest[0],shortest[1],1,shortest[0].getPosition(),pCam.getRotation()]
						else:
							if pD.preGrab[1] == 1:
								o = pD.preGrab[0]
								if o.isValid():
									if hasattr(o,"oId"):
										host.rcon_invoke("Object.active id" + str(o.oId))
										host.rcon_invoke("Object.isVisible 0")
									else: o.setPosition((0.0,83453.0,0.0))
								pD.preGrab[1] = 2
							else:
								o = pD.preGrab[0]
								if o.isValid():
									if hasattr(o,"oId"):
										host.rcon_invoke("Object.active id" + str(o.oId))
										host.rcon_invoke("Object.isVisible 1")
									else: o.setPosition(pD.preGrab[3])
								pD.preGrab[1] = 1
			except:
				pass
		for o in self.flickerObjects:
			try:
				if (o[2] % 2) == 0:
					o[0].setPosition((99999.9,0.0,0.0))
					o[2] -= 1
				else:
					o[0].setPosition(o[1])
					o[2] -= 1
					if o[2] == 0:
						self.flickerObjects.remove(o)
			except:
				ExceptionOutput()

	def freezer(self,data):
		#### Refreeze all frozen objects
		invalidFreezes = []
		for o in self.obj_freeze:
			if self.obj_freeze[o]:
				if (o.isValid()) and (type(self.obj_freeze[o][0]) == type(('tuple','thing'))):
					if o.getPosition() != self.obj_freeze[o][0]:
						o.setPosition(self.obj_freeze[o][0])
					if o.getRotation() != self.obj_freeze[o][1]:
						o.setRotation(self.obj_freeze[o][1])
					try:
						if hasattr(o,"setDamage"):
							if o.getDamage() > 0:
								o.setDamage(10000)
					except:
						pass
				else:
					invalidFreezes.append(o)
		for f in invalidFreezes:
			if self.obj_freeze.has_key(f):
				self.obj_freeze.pop(f)

	def popMessages(self,data):
		for p in bf2.playerManager.getPlayers():
			try:
				pD = self.pdata[p.getName()]
				if pD.popMessage:
					self.sendMsg(p, pD.popMessage, True)
					pD.popMessage = ""
				if pD.saveDelay:
					pD.saveDelay -= 1
				if pD.dupeDelay:
					pD.dupeDelay -= 1
			except:
				try:
					pD.popMessage = ""
				except:
					pass

	def groupCreator(self,data):
		try:
			if self.groupDuper:
				# First, spawn the first object to initate the duplication.
				# [anchor, user, stage, index]
				dAnchor = self.groupDuper[0]
				dUser = self.groupDuper[1]
				try:
					p = bf2.PlayerManager.Player(dUser)
					pD = self.pdata[p.getName()]
				except:
					self.groupDuper = []
					self.groupDuperObjects = []
					self.groupDuperFailures = 0
					self.dbg("Group creator panic, terminating")
					return 0
				dStage = self.groupDuper[2]
				dIndex = self.groupDuper[3]
				dObjects = self.groupDuperObjects
				if (dIndex + 1) > len(dObjects):
					self.groupDuperFailures = 0
					self.dbg("Group created, " + str(dIndex + 1) + " objects processed.")
					self.groupDuper = []
					self.groupDuperObjects = []
					self.onRemoteCommand(dUser, "btnEndGroup")
					return 0
				# First, we run stage 1, which creates the object
				# Then, stage 2 finds the object and adds it to the user's bank of stuff
				obj = dObjects[dIndex]
				if dStage == 1:
					#self.sendMsg(0,str(dIndex))
					#self.dbg("Iterating through object " + str(dIndex))
					if dIndex == 0:
						pPos = p.getDefaultVehicle().getPosition()
						pRot = p.getDefaultVehicle().getRotation()
						pCam = self.getSoldierCam(p)
						pCamRot = pCam.getRotation()
						sPos = self.getVectoredPosition(pPos,pD.tracerDistance,pRot,pCamRot)
						relPos = sPos
						self.groupDuper[0][1] = sPos
					else:
						aPos = dAnchor[1]
						relPos = (obj[1][0] + aPos[0],obj[1][1] + aPos[1],obj[1][2] + aPos[2])
					if pD.objectTypes.count(str(obj[0])) < 1:
						pD.objectTypes.append(str(obj[0]))
					if self.object_types.count(str(obj[0])) < 1:
						self.object_types.append(str(obj[0]))
					sName = obj[0]
					# Attempt to recycle a deleted object
					if not self.allowed_vehicles.count(sName):
						if self.recyclables.has_key(sName):
							if self.recyclables[sName]:
								o = self.recyclables[sName][0]
								if o.isValid():
									self.recyclables[sName].remove(o)
									if self.deleted_objects.count(o):
										self.deleted_objects.remove(o)
									o.setPosition(relPos)
									o.setRotation(obj[2])
									try:
										if hasattr(o,"setDamage"):
											if o.getDamage() > 0:
												o.setDamage(10000)
									except:
										pass
									self.preCompletedObjects.append(o)
									pD.objects.append(o)
									pD.selectedGroup.append(o)
									self.groupDuper[2] = 1
									self.groupDuper[3] += 1
									self.groupDuperFailures = 0
									return
					self.createSpawnerObject(obj[0], relPos, obj[2], -1) # [object name, object position, object rotation]
					self.groupDuper[2] = 2
				elif dStage == 2:
					objectsOfTemplate = bf2.objectManager.getObjectsOfTemplate(obj[0])
					o = None
					o = objectsOfTemplate[-1]
					if o.isValid() and not self.preCompletedObjects.count(o):
						pass
					else:
						o = None
					if o:
						#try:
						#	objs = host.rcon_invoke("object.listObjectsofTemplate " + str(o.templateName))
						#	objs = objs.split()
						#	o.oId = objs[-7]
						#except:
						#	ExceptionOutput()
						#print o.getPosition()
						self.preCompletedObjects.append(o)
						pD.objects.append(o)
						pD.selectedGroup.append(o)
						#self.dbg("Added to selection.")
						self.groupDuper[2] = 1
						self.groupDuper[3] += 1
						self.groupDuperFailures = 0
					else:
						self.groupDuperFailures += 1
					# If unable to find object 4+ times, move on to the next one
					if self.groupDuperFailures > 4:
						self.groupDuper[2] = 1
						self.groupDuper[3] += 1
						self.groupDuperFailures = 0
		except:
			ExceptionOutput()


	#######################################
	#######################################
	#      Sandbox utility functions      #
	#######################################
	#######################################

	def createSpawnerObject(self, sName, sPos, sRot = (0.0,0.0,0.0), user=-1, recycle=False):
		try:
			if recycle and not self.allowed_vehicles.count(sName):
				if self.recyclables.has_key(sName):
					if self.recyclables[sName]:
						o = self.recyclables[sName][0]
						if o.isValid():
							self.recyclables[sName].remove(o)
							if self.deleted_objects.count(o):
								self.deleted_objects.remove(o)
							o.setPosition(sPos)
							o.setRotation(sRot)
							try:
								if hasattr(o,"setDamage"):
									if o.getDamage() > 0:
										o.setDamage(10000)
							except:
								ExceptionOutput()
							self.preCompletedObjects.append(o)
							if user > -1:
								p = bf2.PlayerManager.Player(user)
								pD = self.pdata[p.getName()]
								pD.objects.append(o)
								if pD.autoGrab and ((not self.allowed_vehicles.count(sName)) or (self.autograb_excempt.count(sName))):
								#if pD.autoGrab:
									if pD.grabMode <= 2:
										pD.grabbed = [o,pD.tracerDistance,1,o.getPosition(),o.getRotation()]
									else:
										pPos = p.getDefaultVehicle().getPosition()
										pRot = p.getDefaultVehicle().getRotation()
										pCam = self.getSoldierCam(p)
										pCamRot = pCam.getRotation()
										sPos = self.getVectoredPosition(pPos,pD.tracerDistance,pRot,pCamRot)
										shp = o.getPosition()
										grabOffset = ((sPos[0] - shp[0]),(sPos[1] - shp[1]),(sPos[2] - shp[2]))
										pD.grabbed = [o,pD.tracerDistance,1,o.getPosition(),o.getRotation(),grabOffset]
								if pD.groupCreate:
									pD.selectedGroup.append(o)
									self.dbg("Added object to group")
								if pD.autoLock and self.allow_lock:
									self.obj_locked[o] = p.getName()
									self.dbg("Object auto-locked.")
								#self.dbg("Object successfully recycled.",True)
							return
		except:
			ExceptionOutput()

		if not self.spawnerTemplates.count(sName):
			self.createSpawnerTempl(sName, sName)
			self.spawnerTemplates.append(sName)
		host.rcon_invoke("Object.create sbx_" + sName)
		host.rcon_invoke("Object.absolutePosition " + str(sPos[0]) + "/" + str(sPos[1]) + "/" + str(sPos[2]))
		host.rcon_invoke("Object.rotation " + str(sRot[0]) + "/" + str(sRot[1]) + "/" + str(sRot[2]))
		if (sName == "usaas_stinger_spw" or sName == "aircontroltower") and user:
			p = bf2.PlayerManager.Player(user)
			host.rcon_invoke("Object.team " + str(p.getTeam()))
			host.rcon_invoke("ObjectTemplate.teamOnVehicle 1")
		elif (sName == "usaas_stinger_spw" or sName == "aircontroltower"):
			host.rcon_invoke("Object.team 2")
			host.rcon_invoke("ObjectTemplate.teamOnVehicle 1")
		else:
			host.rcon_invoke("Object.team 2")
		#host.rcon_invoke("Object.delete")
		if (user > -1) and not (self.allowed_vehicles.count(sName) and self.autograb_excempt.count(sName)):
		#if (user > -1):
			self.lastObject = [str(sName),user,sPos,sRot]
			self.lastObjectCounter = 0

	def createSpawnerTempl(self, id, tmpl):
		host.rcon_invoke("ObjectTemplate.create ObjectSpawner sbx_" + id)
		host.rcon_invoke("ObjectTemplate.activeSafe ObjectSpawner sbx_" + id)
		host.rcon_invoke("ObjectTemplate.hasMobilePhysics 0")
		host.rcon_invoke("ObjectTemplate.setObjectTemplate 1 " + tmpl)
		host.rcon_invoke("ObjectTemplate.setObjectTemplate 2 " + tmpl)
		host.rcon_invoke("ObjectTemplate.minSpawnDelay 700000")
		host.rcon_invoke("ObjectTemplate.maxSpawnDelay 900000")

	def createSpawner(self, id, sPos, sRot):
		host.rcon_invoke("Object.create sbx_" + id)
		host.rcon_invoke("Object.absolutePosition " + str(sPos[0]) + "/" + str(sPos[1]) + "/" + str(sPos[2]))
		host.rcon_invoke("Object.rotation " + str(sRot[0]) + "/" + str(sRot[1]) + "/" + str(sRot[2]))
		host.rcon_invoke("Object.team 2")
		host.rcon_invoke("Object.delete")

	def createObject(self, oName, oPos, oRot = (0.0,0.0,0.0), user = 0):
		try:
			oPos = (float(oPos[0]),float(oPos[1]),float(oPos[2]))
			oRot = (float(oRot[0]),float(oRot[1]),float(oRot[2]))
			host.rcon_invoke("gameLogic.createObject " + str(oName))
			self.lastObject = [str(oName),user,oPos,oRot]
			self.lastObjectCounter = 0
		except:
			ExceptionOutput()

	def deleteObject(self, o):
		try:
			if o:
				if o.isValid():
					if hasattr(o,"oId"):
						self.dbg("Deleting using ID")
						host.rcon_invoke("Object.active id" + str(o.oId))
						host.rcon_invoke("Object.delete")
					else:
						o.setPosition((0.0,99999.0,0.0))
						if o.isValid() and o.templateName == "usaas_stinger_spw":
							try:
								o.setDamage(1)
							except:
								pass
						self.deleted_objects.append(o)
						if not self.allowed_vehicles.count(o.templateName):
							if not self.recyclables.has_key(o.templateName):
								self.recyclables[o.templateName] = []
							self.recyclables[o.templateName].append(o)
		except:
			ExceptionOutput()

	def sendMsg(self, p, msg, immediate = False):
		if p:
			if self.pMessages.has_key(msg):
				x = self.pMessages[msg]
				bf2.gameLogic.sendMedalEvent(p, x[0], x[1])
			elif immediate:
				host.rcon_invoke("game.sayAll \"" + "C1001" + str(p.getName()) + " - " + str(msg) + "C1001" + "\"")
			else:
				pD = self.pdata[p.getName()]
				pD.popMessage = str(msg)
		else:
			host.rcon_invoke("game.sayAll \"" + "C1001" + str(msg) + "C1001" + "\"")

	def getSoldierCam(self, player):
		p = getRootParent(player.getDefaultVehicle())
		kids = p.getChildren()
		for child in kids:
			if child.templateName == "SoldierCamera":
				return child
		for child in kids:
			if child.templateName == "SoldierCameraChase":
				return child

	def checkAdmin(self, pName):
		try:
			pName = pName.split(" ")[-1]
			if self.admins.lower().count(pName.lower()):
				return True
			else:
				return False
		except:
			return False


	#######################################
	#######################################
	#      Saving/Loading Creations       #
	#######################################
	#######################################

	def saveSandbox(self, filename):
		file_string = str(host.sgl_getMapName()) + "," + str(self.SANDBOX_VERSION) + "@"
		out_file = open("mods/sandbox/saved/" + str(filename) + ".bf2sbx","w")
		increment = 0
		for t in self.object_types:
			for q in bf2.objectManager.getObjectsOfTemplate(t):
				if q.isValid():
					if not self.deleted_objects.count(q):
						increment += 1
						qPos = q.getPosition()
						qRot = q.getRotation()
						qNam = q.templateName
						file_string += str(qNam) + ";" + str(qPos[0]) + "," + str(qPos[1]) + "," + str(qPos[2]) + ";" + str(qRot[0]) + "," + str(qRot[1]) + "," + str(qRot[2]) + "!"
		out_file.write(file_string)
		out_file.close()
		self.sendMsg(0, "Your sandbox has been saved as " + str(filename) + ".bf2sbx (" + str(increment) + " objects)")
		self.dbg("Saved file, " + str(increment) + " objects")
		if self.ftpUpload and not self.is_spGame:
			result = sandboxUploader.sUpload("mods/sandbox/saved/" + str(filename) + ".bf2sbx",self.ftpOptions)
			if result:
				self.sendMsg(0, "The sandbox has also been uploaded to " + str(self.ftpOptions['dlLocation']))

	def netsaveSandbox(self, filename, p):
		file_string = ""
		increment = 0
		for t in self.object_types:
			for q in bf2.objectManager.getObjectsOfTemplate(t):
				if q.isValid():
					if not self.deleted_objects.count(q):
						increment += 1
						qPos = q.getPosition()
						qRot = q.getRotation()
						qNam = q.templateName
						file_string += str(qNam) + ";" + str(qPos[0]) + "," + str(qPos[1]) + "," + str(qPos[2]) + ";" + str(qRot[0]) + "," + str(qRot[1]) + "," + str(qRot[2]) + "!"
		pName = p.getName()
		pName = pName.split(" ")[-1]
		file_string = pName + "!" + str(p.getProfileId()) + "!" + filename + "!creation!" + str(host.sgl_getMapName()) + "@@" + file_string + "++END"
		result = False
		try:
			result = self.netTalk(file_string,False)
		except:
			ExceptionOutput()
		if result:
			self.sendMsg(p, "Your creation has been saved on the Sandbox Network (" + str(increment) + " objects)")
		else:
			self.sendMsg(p, "Whoops, your creation was not saved.")

	def netloadSandbox(self, filename, p):
		pD = self.pdata[p.getName()]
		netResult = ""
		pName = p.getName()
		pName = pName.split(" ")[-1]
		file_string = pName + "!" + str(p.getProfileId()) + "!" + filename + "!creation" + "++END"
		try:
			netResult = self.netTalk(file_string)
		except:
			ExceptionOutput()
		else:
			try:
				if len(netResult):
					if netResult != "fail":
						self.loadSandbox(netResult, p, "string")
			except:
				ExceptionOutput()

	def loadSandbox(self, filename, p, method="file"):
		pD = self.pdata[p.getName()]
		if method == "file":
			try: in_file = open("mods/sandbox/saved/" + str(filename) + ".bf2sbx","r")
			except:
				self.sendMsg(0, "The file you specified was not found.")
				return 0
			savedStr = in_file.read()
			in_file.close()
		else:
			savedStr = filename
		a = savedStr.split("@")
		w = a[0].split(",")
		if not w.count(self.SANDBOX_VERSION):
			self.sendMsg(0, "This file was created with a different version of BF2:Sandbox. Some objects may be loaded incorrectly.")
		self.sendMsg(0, "Loading sandbox file...this may lag for a little while.")
		if w[0].lower() == str(host.sgl_getMapName()).lower():
			b = a[1].split("!")
			increment = 0
			for objInfo in b:
				if objInfo:
					objInfoList = objInfo.split(";")
					oName = objInfoList[0]
					oPos = objInfoList[1].split(",")
					oRot = objInfoList[2].split(",")
					if self.mp_objects.count(oName.lower()) and not self.is_spGame:
						oName += "_mp"
					self.createSpawnerObject(oName, oPos, oRot)
					if self.object_types.count(oName) < 1:
						self.object_types.append(oName)
					increment += 1
			pD.objectTypes = self.object_types
			if method == "file":
				self.sendMsg(0, str(filename) + ".bf2sbx has been loaded.")
			self.dbg("Loaded file, " + str(increment) + " objects")
		else:
			self.sendMsg(0, "The file you specified was not created in this map.")


	#######################################
	#######################################
	#        Saving/Loading Groups        #
	#######################################
	#######################################

	def saveGroup(self, filename, p):
		file_string = ""
		out_file = open("mods/sandbox/saved/" + str(filename) + ".bf2sbg","w")
		increment = 0
		pD = self.pdata[p.getName()]
		for x in pD.grabbedGroup:
			try:
				if x[0].isValid():
					qPos = x[1]
					qRot = x[2]
					file_string += str(x[0].templateName) + ";" + str(qPos[0]) + "," + str(qPos[1]) + "," + str(qPos[2]) + ";" + str(qRot[0]) + "," + str(qRot[1]) + "," + str(qRot[2]) + "!"
					increment += 1
			except:
				ExceptionOutput()
		out_file.write(file_string)
		out_file.close()
		self.sendMsg(p, "Your group has been saved as " + str(filename) + ".bf2sbg (" + str(increment) + " objects)")

	def netsaveGroup(self, filename, p):
		file_string = ""
		increment = 0
		pD = self.pdata[p.getName()]
		for x in pD.grabbedGroup:
			try:
				if x[0].isValid() and not self.allowed_vehicles.count(x[0].templateName):
					qPos = x[1]
					qRot = x[2]
					file_string += str(x[0].templateName) + ";" + str(qPos[0]) + "," + str(qPos[1]) + "," + str(qPos[2]) + ";" + str(qRot[0]) + "," + str(qRot[1]) + "," + str(qRot[2]) + "!"
					increment += 1
			except:
				ExceptionOutput()
		pName = p.getName()
		pName = pName.split(" ")[-1]
		file_string = pName + "!" + str(p.getProfileId()) + "!" + filename + "!group" + "@@" + file_string + "++END"
		result = False
		try:
			result = self.netTalk(file_string,False)
		except:
			ExceptionOutput()
		if result:
			self.sendMsg(p, "Your group has been saved on the Sandbox Network (" + str(increment) + " objects)")
		else:
			self.sendMsg(p, "Whoops, your group was not saved.")

	def netloadGroup(self, filename, p):
		pD = self.pdata[p.getName()]
		netResult = ""
		pName = p.getName()
		pName = pName.split(" ")[-1]
		file_string = pName + "!" + str(p.getProfileId()) + "!" + filename + "!group" + "++END"
		try:
			netResult = self.netTalk(file_string)
		except:
			ExceptionOutput()
		else:
			try:
				if len(netResult):
					if netResult != "fail":
						self.loadGroup(netResult, p, "string")
			except:
				ExceptionOutput()

	def loadGroup(self, filename, p, method="file"):
		self.dbg("Running loader")
		pD = self.pdata[p.getName()]
		if method == "file":
			try: in_file = open("mods/sandbox/saved/" + str(filename) + ".bf2sbg","r")
			except:
				self.sendMsg(0, "The file you specified was not found.")
				return 0
			savedStr = in_file.read()
			in_file.close()
		elif method == "string":
			savedStr = filename
		b = savedStr.split("!")
		if not self.is_spGame and len(b) > self.groupLoadLimit:
			self.sendMsg(p, "GROUP_LIMITREACHED")
			return
		increment = 0
		self.groupDuperObjects = []
		self.groupDuperObjects.append([0,0,0])
		for objInfo in b:
			if objInfo:
				objInfoList = objInfo.split(";")
				oName = objInfoList[0]
				oPos = objInfoList[1].split(",")
				oRot = objInfoList[2].split(",")
				oPos = (float(oPos[0]),float(oPos[1]),float(oPos[2]))
				oRot = (float(oRot[0]),float(oRot[1]),float(oRot[2]))
				if self.mp_objects.count(oName.lower()) and not self.is_spGame:
					oName += "_mp"
				if oPos == (0.0,0.0,0.0):
					self.groupDuperObjects[0] = [oName,oPos,oRot]
				else:
					self.groupDuperObjects.append([oName,oPos,oRot])
				increment += 1
		pD.selectedGroup = []
		pD.groupCreate = True
		self.groupDuper = [self.groupDuperObjects[0],p.index,1,0]
		if method == "file":
			self.sendMsg(0, str(filename) + ".bf2sbg has been loaded.")
		self.dbg("Loaded group, " + str(increment) + " objects")


	#######################################
	#######################################
	#    Saving/Loading Other Formats     #
	#######################################
	#######################################

	def saveCon(self, saveVehicles=False):
		increment = 0
		object_str = ""
		vehicle_str = ""
		vehicle_spawner_str = ""
		for t in self.object_types:
			for q in bf2.objectManager.getObjectsOfTemplate(t):
				if q.isValid():
					if not self.deleted_objects.count(q):
						increment += 1
						qPos = q.getPosition()
						qRot = q.getRotation()
						qNam = q.templateName
						if not q.isPlayerControlObject:
							object_str += "\nrem *** " + str(qNam) + " ***"
							object_str += "\nObject.create " + str(qNam)
							object_str += "\nObject.absolutePosition " + str(qPos[0]) + "/" + str(qPos[1]) + "/" + str(qPos[2])
							object_str += "\nObject.rotation " + str(qRot[0]) + "/" + str(qRot[1]) + "/" + str(qRot[2])
							object_str += "\nObject.layer 1\n"
						elif q.isPlayerControlObject and saveVehicles:
							vehicle_str += "\nrem [ObjectSpawnerTemplate: sandbox" + str(increment) + "]"
							vehicle_str += "\nObjectTemplate.create ObjectSpawner sandbox" + str(increment)
							vehicle_str += "\nObjectTemplate.activeSafe ObjectSpawner sandbox" + str(increment)
							vehicle_str += "\nObjectTemplate.modifiedByUser elxxrox"
							vehicle_str += "\nObjectTemplate.isNotSaveable 1"
							vehicle_str += "\nObjectTemplate.hasMobilePhysics 0"
							vehicle_str += "\nObjectTemplate.setObjectTemplate 1 " + str(qNam)
							vehicle_str += "\nObjectTemplate.setObjectTemplate 2 " + str(qNam)
							vehicle_str += "\n"
							vehicle_spawner_str += "\n   rem [ObjectSpawner: sandbox" + str(increment) + "]"
							vehicle_spawner_str += "\n   Object.create sandbox" + str(increment)
							vehicle_spawner_str += "\n   Object.absolutePosition " + str(qPos[0]) + "/" + str(qPos[1]) + "/" + str(qPos[2])
							vehicle_spawner_str += "\n   Object.rotation " + str(qRot[0]) + "/" + str(qRot[1]) + "/" + str(qRot[2])
							vehicle_spawner_str += "\n   Object.setControlPointId " + self.con_save_cp
							vehicle_spawner_str += "\n   Object.layer 1\n"
		file = open("mods/sandbox/saved/staticobjects_" + str(host.sgl_getMapName()).lower() + ".txt","w")
		file.write(object_str)
		file.close()
		if saveVehicles:
			file = open("mods/sandbox/saved/gameplayobjectsTemplates_" + str(host.sgl_getMapName()).lower() + ".txt","w")
			file.write(vehicle_str)
			file.close()
			file = open("mods/sandbox/saved/gameplayobjectsSpawners_" + str(host.sgl_getMapName()).lower() + ".txt","w")
			file.write(vehicle_spawner_str)
			file.close()
		self.sendMsg(0, "Saved files for map creation.")

	def dbg(self,msg,override=False):
		if self.enableDebug:
			print msg
			host.rcon_invoke("game.sayAll \"" + str(msg) + "\"")
		elif override:
			host.rcon_invoke("game.sayAll \"" + str(msg) + "\"")


	#######################################
	#######################################
	#           Sandbox Network           #
	#######################################
	#######################################

	def netTalk(self,msg,requestReturn = True):
		sock = socket(AF_INET, SOCK_STREAM)
		self.dbg("Setting to non-blocking")
		sock.settimeout(1.0)
		self.dbg("Connecting to network server")
		try:
			sock.connect((self.netServer, self.netPort))
		except:
			self.dbg("netTalk Error 203 - Could not connect to server")
			ExceptionOutput()
			sock.close()
			return False
		self.dbg("Sending message")
		messlen = 0
		tries = 0
		fail = 0
		data = ""
		while messlen < len(msg):
			tries += 1
			if tries >= 2:
				self.dbg("Failed to send complete message")
				fail = 1
				return False
			try: messlen += sock.send(msg)
			except:
				ExceptionOutput()
			if messlen < len(msg):
				self.dbg("Send looping...")

		# sendMsg("D")
		if requestReturn:
			self.dbg("Getting response")
			data = ""
			#sock.settimeout(0.5)
			while 1:
				try: data += sock.recv(1024)
				except:
					self.dbg("Dead packet")
					ExceptionOutput()
					break
				else:
					self.dbg("Received packet")
					if (data[-3:] == "++E"):
						self.dbg("Manual stoppage")
						data = data[0:-3]
						break
			self.dbg("Checking result")
			if data != "error":
				try:
					#exec(data)
					print data
				except:
					ExceptionOutput()

		self.dbg("Closing socket")
		sock.close()

		if requestReturn:
			return data
		else:
			return True

	#######################################
	#      Battlefield engine events      #
	#######################################
	#######################################

	def onChatMessage(self, playerId, text, channel, flags):
		text = text.replace( "HUD_TEXT_CHAT_TEAM", "" )
		text = text.replace( "HUD_TEXT_CHAT_SQUAD", "" )
		text = text.replace( "HUD_CHAT_DEADPREFIX", "" )
		text = text.replace( "*\xA71DEAD\xA70*", "" )
		if text[0:1] == "@":
			pattern = re.compile(r'@(\w*) ?(.*)')
			matches = pattern.findall(text)
			command = matches[0][0]
			if matches[0][1] != "":
				parameter = matches[0][1]
			else:
				parameter = None
			if playerId == -1: #fix for local servers
				playerId = 255
			p = bf2.PlayerManager.Player(playerId)
			pD = self.pdata[p.getName()]
			self.reloadAdmins()

			if command == "netloadgroup" and parameter:
				if self.groupDuper:
					self.sendMsg(p,"BUSY")
					return 0
				if pD.saveDelay and not self.is_spGame:
					self.sendMsg(p, "SAVEDELAY")
				else:
					self.netloadGroup(parameter,p)
					pD.saveDelay = 200
			if command == "netsavegroup" and parameter:
				if pD.grabbedGroup:
					if pD.saveDelay and not self.is_spGame:
						self.sendMsg(p, "SAVEDELAY")
					else:
						self.netsaveGroup(parameter,p)
						pD.saveDelay = 60
			if command == "version":
				self.sendMsg(0, "This server is running Sandbox " + str(SANDBOX_VERSION) + ", build " + str(SANDBOX_BUILD))

			if self.checkAdmin(p.getName()) or self.is_spGame == 1:
				if command == "netsave" and parameter:
					try:
						self.netsaveSandbox(parameter,p)
					except:
						ExceptionOutput()
				if command == "netload" and parameter:
					self.netloadSandbox(parameter,p)
				if command == "save" and parameter:
					self.saveSandbox(parameter)
				if command == "savegroup" and parameter:
					if pD.grabbedGroup:
						self.saveGroup(parameter,p)
				if command == "loadgroup" and parameter:
					if self.groupDuper:
						self.sendMsg(p,"BUSY")
					else:
						try:
							self.loadGroup(parameter,p)
						except:
							ExceptionOutput()
				if command == "savecon" and parameter:
					self.saveCon(True)
				elif command == "savecon":
					self.saveCon()
				if command == "load" and parameter:
					self.loadSandbox(parameter,p)
				if command == "r" and parameter:
					self.onRemoteCommand(playerId,parameter)
				if command == "c" and parameter:
					host.rcon_invoke(parameter)
				if command == "delall" and parameter:
					self.onRemoteCommand(playerId, "delAll " + str(parameter))
				if command == "dellalln" and parameter:
					self.onRemoteCommand(playerId, "delAllN " + str(parameter))
				if command == "clear":
					host.rcon_invoke("admin.restartmap")

	def onPlayerSpawn(self, player, soldier):
		if self.is_spGame:
			if player.isRemote():
				self.is_spGame = 0

	def onEnterVehicle(self, player, vehicle, freeSoldier = False):
		pD = self.pdata[player.getName()]
		pD.clearGrabbed()

	def onExitVehicle(self, player, vehicle):
		pD = self.pdata[player.getName()]
		pD.clearGrabbed()

	def onPlayerDeathCQ(self, victim, vehicle):
		pD = self.pdata[victim.getName()]
		pD.clearGrabbed()

	def onPlayerKilledCQ(self, victim, attacker, weapon, assists, object):
		pD = self.pdata[victim.getName()]
		pD.clearGrabbed()

	def onPlayerDisconnect(self, player):
		if self.groupDuper:
			if self.groupDuper[1] == player.index:
				self.groupDuper = []
				self.groupDuperObjects = []
		if self.pdata.has_key(player.getName()):
			pD = pdata[player.getName()]
			pD.grabbed = []
			for o in pD.objects:
				self.deleteObject(o)
		for o in self.obj_locked:
			if self.obj_locked[o] == player.getName():
				self.obj_locked.pop(o)
		try:
			self.pdata.pop(player.getName())
		except:
			ExceptionOutput()

	#######################################
	#        Mathematics functions        #
	#######################################
	#######################################

	def getVectoredPosition(self, pPos, distance, pRot, pCamRot):
		from math import pi
		#Calculate the height using a triangle
		p = pCamRot[1] #Pitch of the player's cam
		a = 90.0 - abs(p)
		p = pi * p / 180.0
		a = pi * a / 180.0
		sVz = (distance * math.sin(p))/math.sin(a)
		#convert from crazy-ass BF2 values to polar coordinate values
		new_pRot = -(pRot[0] + pCamRot[0]) + 90.0
		sVx,sVy = self.rect(distance,new_pRot,1) #Convert vector of [spawnDistance] with heading of cam rotation to x,y values
		sPos = (pPos[0] + sVx, pPos[1] - sVz, pPos[2] + sVy)
		return sPos

	def rect(self, r, w, deg=0): # radian if deg=0; degree if deg=1
		from math import cos, sin, pi
		if deg:
			w = pi * w / 180.0
		return r * cos(w), r * sin(w)

	def polar(self, x, y, deg=0): # radian if deg=0; degree if deg=1
		from math import hypot, atan2, pi
		if deg:
			return hypot(x, y), 180.0 * atan2(y, x) / pi
		else:
			return hypot(x, y), atan2(y, x)

	def get2dDistance(self, x, y):
		return math.sqrt(math.pow(y[0] - x[0],2), math.pow(y[1] - x[1],2))

	def get2dRotation(self, x, y, rot):
		print "x=" + str(x) + ",y=" + str(y)
		polar = self.polar(x, y, 1)
		print "polar1 = " + str(polar)
		polar = (polar[0],polar[1] + rot)
		print "polar2 = " + str(polar)
		return self.rect(polar[0],polar[1],1)

	def getVectorDistance(self, pos1, pos2):
		diffVec = [0.0, 0.0, 0.0]
		diffVec[0] = math.fabs(pos1[0] - pos2[0])
		diffVec[1] = math.fabs(pos1[1] - pos2[1])
		diffVec[2] = math.fabs(pos1[2] - pos2[2])
		return math.sqrt(diffVec[0] * diffVec[0] + diffVec[1] * diffVec[1] + diffVec[2] * diffVec[2])

	#######################################
	# Utility functions for other modders #
	#######################################
	#######################################

	def disableBoundaries(self):
		host.rcon_invoke("CombatAreaManager.use 0")
		host.rcon_invoke("CombatAreaManager.timeAllowedOutside 1000.000000")
		self.sBoundaries = False

	def enableBoundaries(self):
		host.rcon_invoke("CombatAreaManager.use 1")
		host.rcon_invoke("CombatAreaManager.timeAllowedOutside 10.00")
		self.sBoundaries = True

	def setAllowedObjects(self,objList):
		if type(self.allowed_objects) == type(objList):
			self.allowed_objects = objList

	def addAllowedObjects(self,obj):
		# This can take either a list or a string.
		if type(self.allowed_objects) == type(obj):
			self.allowed_objects.extend(obj)
		else:
			self.allowed_objects.append(obj)

	def delAllowedObjects(self,obj):
		# This can take either a list or a string.
		if type(self.allowed_objects) == type(obj):
			for x in obj:
				try:
					self.allowed_objects.remove(x)
				except:
					pass #do nothing
		else:
			self.allowed_objects.remove(obj)

	def setAllowLock(self,val):
		if val: self.allow_lock = True
		else: self.allow_lock = False

	def setAllowGrab(self,val):
		if val: self.allow_grab = True
		else: self.allow_grab = False

	def setAllowCreate(self,val,override=True):
		if val: self.allow_create = True
		else: self.allow_create = False
		if override: self.allowCreateOverride = True
		else: self.allowCreateOverride = False

	def setMinPlayers(self,val):
		host.rcon_invoke("sv.numPlayersNeededToStart " + str(val))

def start():
	return sandboxCore()

print "Sandbox core loaded."